import java.io.*;

public class BufferedFileReader extends IOThread
{
	private static final int BFR_BUFFER_COUNT=2;
	private static final int BFR_BUFFER_SIZE=8196;

	public static final byte BFR_CLEAR=(byte)1;
	public static final byte BFR_EXCEPTION=(byte)2;
	public static final byte BFR_EOF=(byte)3;
	public static final byte BFR_STOPPED=(byte)4;

	private InputStream in;
	private Buffer [] work_buffer;
	private Buffer [] avail_buffer;
	private Buffer workOn;
	private int avail,current,work,last_read;
	private byte status;

	public BufferedFileReader(InputStream in,int bufferCount,int bufferSize) throws NullPointerException
	{
		if (in != null) {
			this.in=in;
			if (bufferCount <= 0) {
				bufferCount=BFR_BUFFER_COUNT;;
			}
			if (bufferSize <= 0) {
				bufferSize=BFR_BUFFER_SIZE;
			}
			work_buffer=new Buffer[bufferCount];
			avail_buffer=new Buffer[bufferCount];
			for (work=0; work < bufferCount; ) {
				work_buffer[work++]=new Buffer(bufferSize);
			}
			last_read=0;
			current=avail=0;
			status=BFR_CLEAR;
		} else
			throw new NullPointerException();
	}

	public BufferedFileReader(File file,int bufferCount,int bufferSize) throws NullPointerException, FileNotFoundException
	{
		this(new FileInputStream(file),bufferCount,bufferSize);
	}

	public boolean canRead()
	{
		switch (getStatus()) {
			case BFR_CLEAR:
				return true;
			case BFR_EOF:
				synchronized(avail_buffer) {
					return current < avail;
				}
		}
		return false;
	}

	public boolean canWrite ()
	{
		return getStatus() != BFR_EXCEPTION;
	}

	public byte getStatus()
	{
		synchronized(this) {
			return status;
		}
	}

	public boolean putBuffer(Buffer buffer)
	{
		do {
			if (canWrite()) {
				synchronized(work_buffer) {
					if (work < work_buffer.length) {
						work_buffer[work++]=buffer;
						if (work == 1) {
							return awakeReader();
						}
						return true;
					}
				}
			} else {
				return false;
			}
		} while (waitOutput());
		return false;
	}

	public Buffer getBuffer()
	{
		do {
			if (canRead()) {
				synchronized(avail_buffer) {
					if (current < avail) {
						return avail_buffer[current++ % avail_buffer.length];
					}
				}
			} else {
				return null;
			}
		} while (waitInput());
		return null;
	}

	private boolean waitInput()
	{
		try {
			synchronized(in) {
				in.wait();
			}
		} catch (InterruptedException ie) {
		} catch (IllegalMonitorStateException imse) { }
		return true;
	}
	
	private boolean awakeReader()
	{
		try {
			synchronized(in) {
				in.notifyAll();
			}
			return true;
		} catch (IllegalMonitorStateException imse) { }
		return false;
	}

	private boolean waitOutput()
	{
		try {
			synchronized(this) {
				this.wait();
			}
		} catch (InterruptedException ie) {
		} catch (IllegalMonitorStateException imse) {
			return false;
		}
		return true;
	}
	
	private boolean awakeWriter()
	{
		try {
			synchronized(this) {
				this.notifyAll();
			}
			return true;
		} catch (IllegalMonitorStateException imse) { }
		return false;
	}

	public void setStatus(byte status)
	{
		synchronized(this) {
			this.status=status;
			if (this.isAlive() && status != BFR_CLEAR) {
				awakeReader();
				awakeWriter();
			}
		}
	}

	public void run()
	{
		try {
			do {
				do {
					if (getStatus() == BFR_CLEAR) {
						synchronized(work_buffer) {
							workOn=work > 0 ? work_buffer[--work] : null;
							if (work == work_buffer.length - 1) {
								if (!awakeWriter()) {
									setStatus(BFR_EXCEPTION);
									in.close();
									return;
								}
							}
						}
					} else {
						in.close();
						return;
					}
				} while (workOn == null && waitInput());

				if (workOn == null) {
					setStatus(BFR_EXCEPTION);
					in.close();
					return;
				}

				workOn.size=0;
				while ((last_read=in.read(workOn.buffer,workOn.size,workOn.buffer.length-workOn.size)) >=0
					&& (workOn.size+=last_read) < workOn.buffer.length);

				synchronized(avail_buffer) {
					avail_buffer[avail++ % avail_buffer.length]=workOn;
					if (current == avail - 1) {
						if (!awakeReader()) {
							setStatus(BFR_EXCEPTION);
							in.close();
							return;
						}
					}
				}
			} while (last_read != -1);
			setStatus(awakeReader() ? BFR_EOF : BFR_EXCEPTION);
			in.close();
		} catch (Exception e) {
			setStatus(BFR_EXCEPTION);
		}
	}
}
