import java.io.*;

public class BufferedFileWriter extends IOThread
{
	private static final int BFW_BUFFER_COUNT=2;
	private static final int BFW_BUFFER_SIZE=8196;

	public static final byte BFW_CLEAR=(byte)1;
	public static final byte BFW_EXCEPTION=(byte)2;
	public static final byte BFW_EOF=(byte)3;
	public static final byte BFW_STOPPED=(byte)4;

	private OutputStream out;
	private Buffer [] work_buffer;
	private Buffer [] avail_buffer;
	private Buffer workOn;
	private int avail,current,work;
	private byte status;

	public BufferedFileWriter(OutputStream out,int bufferCount,int bufferSize) throws NullPointerException
	{
		if (out != null) {
			this.out=out;
			if (bufferCount <= 0) {
				bufferCount=BFW_BUFFER_COUNT;;
			}
			if (bufferSize <= 0) {
				bufferSize=BFW_BUFFER_SIZE;
			}
			work_buffer=new Buffer[bufferCount];
			avail_buffer=new Buffer[bufferCount];
			for (avail=0; avail < bufferCount; ) {
				avail_buffer[avail++]=new Buffer(bufferSize);
			}
			current=work=0;
			status=BFW_CLEAR;
		} else
			throw new NullPointerException();
	}

	public BufferedFileWriter(File file,int bufferCount,int bufferSize) throws NullPointerException, FileNotFoundException
	{
		this(new FileOutputStream(file),bufferCount,bufferSize);
	}

	public boolean canRead()
	{
		switch (getStatus()) {
			case BFW_CLEAR:
				return true;
			case BFW_EOF:
				synchronized(work_buffer) {
					return current < work;
				}
		}
		return false;
	}

	public boolean canWrite ()
	{
		return getStatus() != BFW_EXCEPTION;
	}

	public byte getStatus()
	{
		synchronized(this) {
			return status;
		}
	}

	public boolean putBuffer(Buffer buffer)
	{
		do {
			if (canWrite()) {
				synchronized(work_buffer) {
					if (work - current < work_buffer.length) {
						work_buffer[work++ % work_buffer.length]=buffer;
						if (work - current == 1) {
							return awakeReader();
						}
						return true;
					}
				}
			} else {
				return false;
			}
		} while (waitOutput());
		return false;
	}

	public Buffer getBuffer()
	{
		do {
			if (canRead()) {
				synchronized(avail_buffer) {
					if (avail > 0) {
						return avail_buffer[--avail];
					}
				}
			} else {
				return null;
			}
		} while (waitInput());
		return null;
	}

	private boolean waitInput()
	{
		try {
			synchronized(out) {
				out.wait();
			}
		} catch (InterruptedException ie) {
		} catch (IllegalMonitorStateException imse) { }
		return true;
	}
	
	private boolean awakeReader()
	{
		try {
			synchronized(out) {
				out.notifyAll();
			}
			return true;
		} catch (IllegalMonitorStateException imse) { }
		return false;
	}

	private boolean waitOutput()
	{
		try {
			synchronized(this) {
				this.wait();
			}
		} catch (InterruptedException ie) {
		} catch (IllegalMonitorStateException imse) { }
		return true;
	}
	
	private boolean awakeWriter()
	{
		try {
			synchronized(this) {
				this.notifyAll();
			}
			return true;
		} catch (IllegalMonitorStateException imse) { }
		return false;
	}

	public void setStatus(byte status)
	{
		synchronized(this) {
			this.status=status;
			if (this.isAlive() && status != BFW_CLEAR) {
				awakeReader();
				awakeWriter();
			}
		}
	}

	public void run()
	{
		try {
			boolean notExit;
			do {
				workOn=null;
				do {
					if (getStatus() == BFW_CLEAR) {
						synchronized(work_buffer) {
							if (current < work) {
								workOn=work_buffer[current++ % work_buffer.length];
							}
						}
					} else {
						out.close();
						return;
					}
				} while (workOn == null && waitInput());
				if (workOn == null) {
					setStatus(BFW_EXCEPTION);
					out.close();
					return;
				}
				out.write(workOn.buffer,0,workOn.size);
				notExit=workOn.size > 0;
				workOn.size=0;
				synchronized(avail_buffer) {
					avail_buffer[avail++]=workOn;
					if (!awakeReader()) {
						setStatus(BFW_EXCEPTION);
						out.close();
						return;
					}
				}
			} while (notExit);
			setStatus(awakeReader() ? BFW_EOF : BFW_EXCEPTION);
			out.close();
		} catch (Exception e) {
			setStatus(BFW_EXCEPTION);
		}
	}
}
